(ns 
  ^{:author "Brian Craft"
    :doc "Shim for old jvms without file watch service."}
  cavm.dir-watch
  (:require [clojure.tools.logging :refer [warn]])
  (:require [clojure.data.priority-map :refer [priority-map]])
  (:import [java.io File]))

(defn type-file [dir]
  (filter #(not (.isDirectory ^File %)) (file-seq dir)))

(defn file-timestamps [dir]
  (let [files (type-file dir)]
    (into {} (apply map vector [files (map #(.lastModified ^File %) files)]))))

(defn diff
  "Diff two maps. This is much faster than using clojure.set methods."
  [a b]
  (let [old-keys (filter #(not (b %)) (keys a))
        new-keys (filter #(not (a %)) (keys b))
        changed-keys (filter #(when-let [v (b %)] (not (= v (a %)))) (keys a))]
    {:old old-keys :new new-keys :changed changed-keys}))

(defn check-files
  "Return new, old, and changed files, plus latest file timestamps."
  [dir old-times]
  (let [new-times (file-timestamps dir)]
    (assoc (diff old-times new-times) :files new-times)))

(defn poller [running poll-agent cb interval dir old]
  (Thread/sleep interval)
  (when @running
    (let [changes (check-files dir old)]
      (when (some #(> (count %) 0) (map changes [:old :new :changed]))
        (cb changes))
      (send-off poll-agent (partial poller running poll-agent cb interval dir))
      (:files changes))))

(def sec 1000)

(defn- log-error [a e]
  (warn e "File watcher"))

(defn dir-watcher
  "Watch a directory tree 'dir' for changes, calling 'cb' on any change, with
  a map of :old :new and :changed file lists, plus a map of all files to their
  timestamps. Returns a function that terminates the watcher."
  [cb dir & [{:keys [interval initial-files]
              :or {interval (* 30 sec) initial-files nil}}]]
  (let [file-list (agent (or initial-files (file-timestamps dir)))
        running (atom true)]
    (set-error-handler! file-list log-error)
    (send-off file-list (partial poller running file-list cb interval dir))
    (fn [] (reset! running false))))

(defn dir-change-queue
  "Accumulate filesystem changes into the given atom, as {file timestamp ...}.
  This should typically be a priority queue."
  [dir queue & [{:keys [interval]
                                      :or {interval (* 30 sec)}
                                      :as opts}]]
  (dir-watcher (fn [{:keys [new old changed files]}]
                 (swap! queue into (select-keys files (concat new changed))))
               dir
               opts))

; Handle case where file is put on queue, but then deleted. Does
; it stay on queue? Loader should skip it.

;
; The timestamp should be recorded just before computing the hash.
;
; Use a priority queue, file -> lastModified. Pull the oldest
; lastModified, at least as old as epsilon. The loader agent should
; - take item off the queue
; - read lastModified
; - if it's newer than epsilon, put it back on the queue
; - otherwise add to 'reference' map, compute file hash & load
; - on error, remove from 'reference' map

; Reference map: do we need it? Or should we rely on the db cache?

;
; Files enter the queue by having timestamp different from 'reference'.
; What about load failures? We don't want them to repeat forever. Only
; retry on startup?

; At startup, get 'reference' from db, before starting loader. Get
; all current timestamps & compute difference, putting all changes on
; queue.

; With OS support, listen to changed, deleted, created, and put any
; changes on the queue.

; Without OS support, keep full list of previous file timestamps &
; diff with current. This is the algorithm we have above.

; Loader runs when queue atom is modified. Or is the queue an agent,
; and we send the changes to the agent?

; When does loader run if nothing on the queue is ready to run? Should
; check queue at least after (epsilon - oldest). What if a bunch of files
; are updated? We don't want to queue a run for every one. What if a file
; is mv'ed and has an old timestamp?

; async tasks
; - accumulating filesystem change
; - loading files



; The OS support should give us a list of files changed, deleted, created, probably
; from when the service is started. It doesn't have the initial state. We need
; to handle that elsewhere.

; File change service: files changed since service start
; Queue manager: take initial state from db, listen to file change, user requests,
;                send to loader.
; Loader: loads whatever is sent to it, subject to hash check
;               -> what updates the file hash cache???

; We could pass the loader an atom, where it would update file timestamps as
; they are loaded, delete them on failure. How would we keep the queue list up
; to date? Could also have the loader update that?

; How can the quiet period work? The dir-watcher will not flag it again
; if it doesn't change, same as with OS support. queue must maintain the
; map of loaded files.

; On notification we update timestamps

; What we really need is a full map of files -> timestamps on the filesystem,
; to diff with the db. So it doesn't make sense to write dir-watcher to mimick
; the OS support. We should wrap the OS support to create the map.
; For the OS support we should register for changes, start building the map.
; how to reconcile the two? Need an agent for the async. Is there any way to
; guarantee no ordering problems?
; 1 - we read timestamp
; 2 - file changes
; 3 - we get notification, update agent
; 4 - we send timestamp to agent
; Instead, the agent always calculates the timestamp. That serializes the calls.

; Goes into queue module:
;(defn queue [files]
;  (println "queuing" files))

;        now (.getTime (java.util.Date.))
;        to-load (concat new (filter #(> quiet-period (- now (new-times %)))))]

